# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_db import exception as db_exc
from oslo_log import log as logging
from oslo_utils import excutils
from oslo_utils import timeutils
from oslo_utils import uuidutils
from sqlalchemy import exc as sql_exc
from sqlalchemy.orm import exc

from neutron.db import common_db_mixin

from networking_cisco.backwards_compatibility import (cb_constants as
    svc_constants)
from networking_cisco.plugins.cisco.db.device_manager import hd_models
from networking_cisco.plugins.cisco.extensions import ciscohostingdevicemanager

LOG = logging.getLogger(__name__)

AUTO_DELETE_DEFAULT = ciscohostingdevicemanager.AUTO_DELETE_DEFAULT


class HostingDeviceDBMixin(
        ciscohostingdevicemanager.CiscoHostingDevicePluginBase,
        common_db_mixin.CommonDbMixin):
    """A class implementing DB functionality for hosting devices."""

    def create_hosting_device(self, context, hosting_device):
        LOG.debug("create_hosting_device() called")
        hd = hosting_device['hosting_device']
        with context.session.begin(subtransactions=True):
            credentials_id = hd.get('credentials_id')
            if credentials_id is None:
                hdt_db = self._get_hosting_device_template(context,
                                                           hd['template_id'])
                credentials_id = hdt_db['default_credentials_id']
            hd_db = hd_models.HostingDevice(
                id=self._get_id(hd),
                complementary_id=hd.get('complementary_id'),
                tenant_id=hd['tenant_id'],
                template_id=hd['template_id'],
                credentials_id=credentials_id,
                name=hd.get('name'),
                description=hd.get('description'),
                device_id=hd.get('device_id'),
                admin_state_up=hd.get('admin_state_up', True),
                management_ip_address=hd['management_ip_address'],
                management_port_id=hd['management_port_id'],
                protocol_port=hd.get('protocol_port'),
                cfg_agent_id=hd.get('cfg_agent_id'),
                created_at=hd.get('created_at', timeutils.utcnow()),
                status=hd.get('status', svc_constants.ACTIVE),
                tenant_bound=hd.get('tenant_bound'),
                auto_delete=hd.get('auto_delete', AUTO_DELETE_DEFAULT))
            context.session.add(hd_db)
        return self._make_hosting_device_dict(hd_db)

    def update_hosting_device(self, context, id, hosting_device):
        LOG.debug("update_hosting_device() called")
        hd = hosting_device['hosting_device']
        with context.session.begin(subtransactions=True):
            #TODO(bobmel): handle tenant_bound changes
            hd_query = context.session.query(hd_models.HostingDevice)
            if not hd_query.filter_by(id=id).update(hd):
                raise ciscohostingdevicemanager.HostingDeviceNotFound(id=id)
            #TODO(bobmel): notify_agent on changes to credentials,
            # admin_state_up, tenant_bound
        return self.get_hosting_device(context, id)

    def delete_hosting_device(self, context, id):
        LOG.debug("delete_hosting_device() called")
        try:
            with context.session.begin(subtransactions=True):
                hd_query = context.session.query(hd_models.HostingDevice)
                if not hd_query.filter_by(id=id).delete():
                    raise ciscohostingdevicemanager.HostingDeviceNotFound(
                        id=id)
        except db_exc.DBError as e:
            with excutils.save_and_reraise_exception() as ctxt:
                if isinstance(e.inner_exception, sql_exc.IntegrityError):
                    ctxt.reraise = False
                    raise ciscohostingdevicemanager.HostingDeviceInUse(id=id)

    def get_hosting_device(self, context, id, fields=None):
        LOG.debug("get_hosting_device() called")
        hd_db = self._get_hosting_device(context, id)
        return self._make_hosting_device_dict(hd_db)

    def get_hosting_devices(self, context, filters=None, fields=None,
                            sorts=None, limit=None, marker=None,
                            page_reverse=False):
        LOG.debug("get_hosting_devices() called")
        return self._get_collection(context, hd_models.HostingDevice,
                                    self._make_hosting_device_dict,
                                    filters=filters, fields=fields,
                                    sorts=sorts, limit=limit,
                                    marker_obj=marker,
                                    page_reverse=page_reverse)

    def get_hosting_devices_db(self, context, filters=None, sorts=None,
                               limit=None, marker=None, page_reverse=False):
        LOG.debug("get_hosting_devices_db() called")
        return self._get_collection_query(
            context, hd_models.HostingDevice, filters=filters, sorts=sorts,
            limit=limit, marker_obj=marker, page_reverse=page_reverse)

    def create_hosting_device_template(self, context, hosting_device_template):
        LOG.debug("create_hosting_device_template() called")
        hdt = hosting_device_template['hosting_device_template']
        #TODO(bobmel): check service types
        with context.session.begin(subtransactions=True):
            hdt_db = hd_models.HostingDeviceTemplate(
                id=self._get_id(hdt),
                tenant_id=hdt['tenant_id'],
                name=hdt.get('name'),
                enabled=hdt.get('enabled', True),
                host_category=hdt['host_category'],
                service_types=hdt.get('service_types'),
                image=hdt.get('image'),
                flavor=hdt.get('flavor'),
                default_credentials_id=hdt.get('default_credentials_id'),
                configuration_mechanism=hdt.get('configuration_mechanism'),
                protocol_port=hdt.get('protocol_port'),
                booting_time=hdt.get('booting_time'),
                slot_capacity=hdt['slot_capacity'],
                desired_slots_free=hdt['desired_slots_free'],
                tenant_bound=':'.join(hdt['tenant_bound']),
                device_driver=hdt['device_driver'],
                plugging_driver=hdt['plugging_driver'])
            context.session.add(hdt_db)
        return self._make_hosting_device_template_dict(hdt_db)

    def update_hosting_device_template(self, context,
                                       id, hosting_device_template):
        LOG.debug("update_hosting_device_template() called")
        hdt = hosting_device_template['hosting_device_template']
        tenant_bound = hdt.get('tenant_bound')
        if tenant_bound is not None:
            hdt['tenant_bound'] = ':'.join(tenant_bound)
        with context.session.begin(subtransactions=True):
            hdt_query = context.session.query(hd_models.HostingDeviceTemplate)
            if not hdt_query.filter_by(id=id).update(hdt):
                raise ciscohostingdevicemanager.HostingDeviceTemplateNotFound(
                    id=id)
        return self.get_hosting_device_template(context, id)

    def delete_hosting_device_template(self, context, id):
        LOG.debug("delete_hosting_device_template() called")
        try:
            with context.session.begin(subtransactions=True):
                hdt_query = context.session.query(
                    hd_models.HostingDeviceTemplate)
                if not hdt_query.filter_by(id=id).delete():
                    raise (ciscohostingdevicemanager.
                        HostingDeviceTemplateNotFound(id=id))
        except db_exc.DBError as e:
            with excutils.save_and_reraise_exception() as ctxt:
                if isinstance(e.inner_exception, sql_exc.IntegrityError):
                    ctxt.reraise = False
                    raise (ciscohostingdevicemanager.
                           HostingDeviceTemplateInUse(id=id))

    def get_hosting_device_template(self, context, id, fields=None):
        LOG.debug("get_hosting_device_template() called")
        hdt_db = self._get_hosting_device_template(context, id)
        return self._make_hosting_device_template_dict(hdt_db)

    def get_hosting_device_templates(self, context, filters=None, fields=None,
                                     sorts=None, limit=None, marker=None,
                                     page_reverse=False):
        LOG.debug("get_hosting_device_templates() called")
        return self._get_collection(context, hd_models.HostingDeviceTemplate,
                                    self._make_hosting_device_template_dict,
                                    filters=filters, fields=fields,
                                    sorts=sorts, limit=limit,
                                    marker_obj=marker,
                                    page_reverse=page_reverse)

    def _get_id(self, res):
        uuid = res.get('id')
        if uuid:
            return uuid
        return uuidutils.generate_uuid()

    def _get_hosting_device(self, context, id):
        try:
            return self._get_by_id(context, hd_models.HostingDevice, id)
        except exc.NoResultFound:
            raise ciscohostingdevicemanager.HostingDeviceNotFound(id=id)

    def _make_hosting_device_dict(self, hd, fields=None):
        res = {'id': hd['id'],
               'complementary_id': hd['complementary_id'],
               'tenant_id': hd['tenant_id'],
               'template_id': hd['template_id'],
               'credentials_id': hd['credentials_id'],
               'name': hd['name'],
               'description': hd['description'],
               'device_id': hd['device_id'],
               'admin_state_up': hd['admin_state_up'],
               'management_ip_address': hd['management_ip_address'],
               'management_port_id': hd['management_port_id'],
               'protocol_port': hd['protocol_port'],
               'cfg_agent_id': hd['cfg_agent_id'],
               'created_at': hd['created_at'],
               'status': hd['status'],
               'tenant_bound': hd['tenant_bound'],
               'auto_delete': hd['auto_delete']}
        return self._fields(res, fields)

    def _get_hosting_device_template(self, context, id):
        try:
            return self._get_by_id(context, hd_models.HostingDeviceTemplate,
                                   id)
        except exc.NoResultFound:
            raise ciscohostingdevicemanager.HostingDeviceTemplateNotFound(
                id=id)

    def _make_hosting_device_template_dict(self, hdt, fields=None):
        tb = hdt['tenant_bound'].split(':') if len(hdt['tenant_bound']) else []
        res = {'id': hdt['id'],
               'tenant_id': hdt['tenant_id'],
               'name': hdt['name'],
               'enabled': hdt['enabled'],
               'host_category': hdt['host_category'],
               'service_types': hdt['service_types'],
               'image': hdt['image'],
               'flavor': hdt['flavor'],
               'default_credentials_id': hdt['default_credentials_id'],
               'configuration_mechanism': hdt['configuration_mechanism'],
               'protocol_port': hdt['protocol_port'],
               'booting_time': hdt['booting_time'],
               'slot_capacity': hdt['slot_capacity'],
               'desired_slots_free': hdt['desired_slots_free'],
               'tenant_bound': tb,
               'device_driver': hdt['device_driver'],
               'plugging_driver': hdt['plugging_driver']}
        return self._fields(res, fields)



# Copyright 2011 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import constants


# Constants related to device manager
# ===================================
# Service type for device manager plugin
DEVICE_MANAGER = "DEVICE_MANAGER"

# RPC topic for device manager
DEVICE_MANAGER_PLUGIN = 'n-devmgr-plugin'


# Type and topic for Cisco cfg agent
# ==================================
AGENT_TYPE_CFG = 'Cisco cfg agent'
AGENT_TYPE_L3_CFG = 'Cisco l3 cfg agent'

# Topic for Cisco configuration agent
CFG_AGENT = 'cisco_cfg_agent'
# Topic for routing service helper in Cisco configuration agent
CFG_AGENT_L3_ROUTING = 'cisco_cfg_agent_l3_routing'

# Device manager and (hosting) devices
# ====================================
# Default name of hosting device template for network nodes
# hosting Linux network namespace-based Neutron routers.
NETWORK_NODE_TEMPLATE = 'NetworkNode_template'

# Statuses of hosting devices
# ===========================
# Active means the hosting device is up, responds to pings and is configurable
HD_ACTIVE = 'ACTIVE'
# Not responding means that hosting device does not respond to pings but has
# not yet been determined to be dead or faulty
HD_NOT_RESPONDING = 'NOT RESPONDING'
# Error means that hosting device has been determined to be faulty, meaning it
# may respond to pings but other symptoms indicate it is faulty
HD_ERROR = 'ERROR'
# Dead means that hosting device has been determined to be dead in that it
# does not respond to pings even given multiple, repeated attempts
HD_DEAD = 'DEAD'


# Routing service
# ==============
# Default name of router type for Neutron routers implemented
# as Linux network namespaces in network nodes.
NAMESPACE_ROUTER_TYPE = 'Namespace_Neutron_router'

# Default name of router type for Neutron routers implemented
# as VRFs inside a ASR1k physical device.
ASR1K_ROUTER_TYPE = 'ASR1k_router'

# Router roles
ROUTER_ROLE_GLOBAL = 'Global'
ROUTER_ROLE_LOGICAL_GLOBAL = 'Logical-Global'
ROUTER_ROLE_HA_REDUNDANCY = 'HA-Redundancy'
ALLOWED_ROUTER_ROLES = [ROUTER_ROLE_GLOBAL, ROUTER_ROLE_LOGICAL_GLOBAL,
                        ROUTER_ROLE_HA_REDUNDANCY]

# Prefix of name given to global routers
ROUTER_ROLE_NAME_PREFIX = ROUTER_ROLE_GLOBAL + '-router'
LOGICAL_ROUTER_ROLE_NAME = ROUTER_ROLE_LOGICAL_GLOBAL + '-router'
# Number of characters of hosting device id added to role name
ROLE_ID_LEN = 17

# The status of a Neutron Router created using the
# Cisco service plugin is one of the following:
# Created but not scheduled nor deployed
ROUTER_PENDING_CREATE = "PENDING_CREATE"
#  Scheduling in progress
ROUTER_SCHEDULING = 'SCHEDULING'
# Backlogged due to unsuccessful scheduling attempt
ROUTER_BACKLOGGED = 'BACKLOGGED'
# Backlogged due to non-ready hosting device (e.g., still booting)
ROUTER_WAITING_HOST = 'AWAITING_HOST'
# Deployed and configured
ROUTER_ACTIVE = "ACTIVE"
# Deletion in progress (by cfg agent)
ROUTER_PENDING_DELETE = "PENDING_DELETE"
# Some db states in flight so all info is not yet available
ROUTER_INFO_INCOMPLETE = "INFO_INCOMPLETE"
# Values for network profile fields
ADD_TENANTS = 'add_tenants'
REMOVE_TENANTS = 'remove_tenants'

# Owner identifier for auxiliary gateway ports of global routers
#DEVICE_OWNER_GLOBAL_ROUTER_GW = (constants.DEVICE_OWNER_NETWORK_PREFIX +
#                                 "global_router_gateway")
DEVICE_OWNER_GLOBAL_ROUTER_GW = constants.DEVICE_OWNER_ROUTER_INTF
AUXILIARY_GATEWAY_KEY = constants.INTERFACE_KEY

# Dict key for above gateway ports in global router dicts
GATEWAY_KEY = '_gateways'



# Copyright 2017 Cisco Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from neutron.db import common_db_mixin
from neutron_lib.api import validators

from networking_cisco import backwards_compatibility as bc
from networking_cisco.backwards_compatibility import ml2_api as api

LOG = logging.getLogger(__name__)


class CiscoProviderNetDriver(api.ExtensionDriver,
                             common_db_mixin.CommonDbMixin):
    _supported_extension_alias = 'provider'

    def initialize(self):
        LOG.info("CiscoProviderNetDriver initialization complete")

    @property
    def extension_alias(self):
        return self._supported_extension_alias

    def process_create_network(self, context, data, result):
        if validators.is_attr_set(data.get(bc.providernet.SEGMENTATION_ID)):
            result['is_provider_network'] = True



# Copyright 2017 Cisco Systems.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""
This file contains the implementation of Topology Discovery of servers and
their associated leaf switches using Open source implementation of LLDP.
www.open-lldp.org
"""

from networking_cisco.apps.saf.agent.topo_disc import (
    topo_disc_constants as constants)
from networking_cisco.apps.saf.agent.topo_disc import pub_lldp_api as pub_lldp
from networking_cisco.apps.saf.common import dfa_logger as logging
from networking_cisco.apps.saf.common import dfa_sys_lib as sys_utils
from networking_cisco.apps.saf.common import utils

LOG = logging.getLogger(__name__)


class TopoIntfAttr(object):

    """Class that stores the interface attributes. """

    def __init__(self, protocol_interface, phy_interface):
        """Class Init. """
        self.init_params(protocol_interface, phy_interface)

    def init_params(self, protocol_interface, phy_interface):
        """Initializing parameters. """
        self.lldp_cfgd = False
        self.local_intf = protocol_interface
        self.phy_interface = phy_interface
        self.remote_evb_cfgd = False
        self.remote_evb_mode = None
        self.remote_mgmt_addr = None
        self.remote_system_desc = None
        self.remote_system_name = None
        self.remote_port = None
        self.remote_chassis_id_mac = None
        self.remote_port_id_mac = None
        self.local_evb_cfgd = False
        self.local_evb_mode = None
        self.local_mgmt_address = None
        self.local_system_desc = None
        self.local_system_name = None
        self.local_port = None
        self.local_chassis_id_mac = None
        self.local_port_id_mac = None
        self.db_retry_status = False
        self.topo_send_cnt = 0
        self.bond_interface = None
        self.bond_member_ports = None

    def update_lldp_status(self, status):
        """Update the LLDP cfg status. """
        self.lldp_cfgd = status

    def cmp_update_bond_intf(self, bond_interface):
        """Update the bond interface and its members.

        Update the bond interface, if this interface is a part of bond
        Return True if there's a change.
        """
        if bond_interface != self.bond_interface:
            self.bond_interface = bond_interface
            self.bond_member_ports = sys_utils.get_member_ports(bond_interface)
            return True
        return False

    def get_lldp_status(self):
        """Retrieve the LLDP cfg status. """
        return self.lldp_cfgd

    def get_db_retry_status(self):
        """Retrieve the RPC retru status.

        This retrieves the number of times RPC was retried with the server.
        """
        return self.db_retry_status

    def get_phy_interface(self):
        """Retrieves the physical interface. """
        return self.phy_interface

    def store_db_retry_status(self, status):
        """This stores the number of times RPC was retried with the server. """
        self.db_retry_status = status

    def get_topo_disc_send_cnt(self):
        """Retrieve the topology status send count for this interface. """
        return self.topo_send_cnt

    def incr_topo_disc_send_cnt(self):
        """Increment the topology status send count for this interface. """
        self.topo_send_cnt += 1

    def reset_topo_disc_send_cnt(self):
        """Reset the topology status send count for this interface. """
        self.topo_send_cnt = 0

    def remote_evb_mode_uneq_store(self, remote_evb_mode):
        """Saves the EVB mode, if it is not the same as stored. """
        if remote_evb_mode != self.remote_evb_mode:
            self.remote_evb_mode = remote_evb_mode
            return True
        return False

    def remote_evb_cfgd_uneq_store(self, remote_evb_cfgd):
        """This saves the EVB cfg, if it is not the same as stored. """
        if remote_evb_cfgd != self.remote_evb_cfgd:
            self.remote_evb_cfgd = remote_evb_cfgd
            return True
        return False

    def remote_mgmt_addr_uneq_store(self, remote_mgmt_addr):
        """This function saves the MGMT address, if different from stored. """
        if remote_mgmt_addr != self.remote_mgmt_addr:
            self.remote_mgmt_addr = remote_mgmt_addr
            return True
        return False

    def remote_sys_desc_uneq_store(self, remote_system_desc):
        """This function saves the system desc, if different from stored. """
        if remote_system_desc != self.remote_system_desc:
            self.remote_system_desc = remote_system_desc
            return True
        return False

    def remote_sys_name_uneq_store(self, remote_system_name):
        """This function saves the system name, if different from stored. """
        if remote_system_name != self.remote_system_name:
            self.remote_system_name = remote_system_name
            return True
        return False

    def remote_port_uneq_store(self, remote_port):
        """This function saves the port, if different from stored. """
        if remote_port != self.remote_port:
            self.remote_port = remote_port
            return True
        return False

    def remote_chassis_id_mac_uneq_store(self, remote_chassis_id_mac):
        """This function saves the Chassis MAC, if different from stored. """
        if remote_chassis_id_mac != self.remote_chassis_id_mac:
            self.remote_chassis_id_mac = remote_chassis_id_mac
            return True
        return False

    def remote_port_id_mac_uneq_store(self, remote_port_id_mac):
        """This function saves the port MAC, if different from stored. """
        if remote_port_id_mac != self.remote_port_id_mac:
            self.remote_port_id_mac = remote_port_id_mac
            return True
        return False


class TopoDiscPubApi(object):
    topo_intf_obj_dict = {}

    @classmethod
    def store_obj(cls, intf, obj):
        """Stores the topo object. """
        cls.topo_intf_obj_dict[intf] = obj

    @classmethod
    def get_lldp_status(cls, intf):
        """Retrieves the LLDP status. """
        if intf not in cls.topo_intf_obj_dict:
            LOG.error("Interface %s not configured at all", intf)
            return False
        intf_obj = cls.topo_intf_obj_dict.get(intf)
        return intf_obj.get_lldp_status()


class TopoDisc(TopoDiscPubApi):

    """Topology Discovery Top level class once. """

    def __init__(self, cb, root_helper, intf_list=None, all_intf=True):
        """Initialization routine, to configure interface.

        Also create the periodic task.
        cb => Callback in case any of the interface TLV changes.
        intf_list => List of interfaces to be LLDP enabled and monitored.
        all_intf => Boolean that signifies if all physical interfaces are to
        be monitored. intf_list will be None, if this variable is True.
        """
        self.pub_lldp = pub_lldp.LldpApi(root_helper)
        self._init_cfg_interfaces(cb, intf_list, all_intf)
        per_task = utils.PeriodicTask(constants.PERIODIC_TASK_INTERVAL,
                                      self.periodic_discovery_task)
        per_task.run()

    def _init_cfg_interfaces(self, cb, intf_list=None, all_intf=True):
        """Configure the interfaces during init time. """
        if not all_intf:
            self.intf_list = intf_list
        else:
            self.intf_list = sys_utils.get_all_run_phy_intf()
        self.cb = cb
        self.intf_attr = {}
        self.cfg_lldp_interface_list(self.intf_list)

    def cfg_intf(self, protocol_interface, phy_interface=None):
        """Called by application to add an interface to the list. """
        self.intf_list.append(protocol_interface)
        self.cfg_lldp_interface(protocol_interface, phy_interface)

    def uncfg_intf(self, intf):
        """Called by application to remove an interface to the list.

        From an applications perspective, it makes sense to have this function.
        But, here no action can be taken for the following reasons, but just
        having it as a place-holder for tomorrow.
        => Can't remove interface from the list since DB in server may appear
        stale. self.intf_list.remove(intf)
        => One can just remove the interface DB, but need to retry that till
        it succeeds, so it has to be in periodic loop.
        => So, currently leaving it as is, since LLDP frames won't be obtained
        over the bridge, the periodic handler will automatically remove the
        DB for this interface from server
        """
        pass

    def create_attr_obj(self, protocol_interface, phy_interface):
        """Creates the local interface attribute object and stores it. """
        self.intf_attr[protocol_interface] = TopoIntfAttr(
            protocol_interface, phy_interface)
        self.store_obj(protocol_interface, self.intf_attr[protocol_interface])

    def get_attr_obj(self, intf):
        """Retrieve the interface object. """
        return self.intf_attr[intf]

    def cmp_store_tlv_params(self, intf, tlv_data):
        """Compare and store the received TLV.

        Compares the received TLV with stored TLV. Store the new TLV if it is
        different.
        """
        flag = False
        attr_obj = self.get_attr_obj(intf)
        remote_evb_mode = self.pub_lldp.get_remote_evb_mode(tlv_data)
        if attr_obj.remote_evb_mode_uneq_store(remote_evb_mode):
            flag = True
        remote_evb_cfgd = self.pub_lldp.get_remote_evb_cfgd(tlv_data)
        if attr_obj.remote_evb_cfgd_uneq_store(remote_evb_cfgd):
            flag = True
        remote_mgmt_addr = self.pub_lldp.get_remote_mgmt_addr(tlv_data)
        if attr_obj.remote_mgmt_addr_uneq_store(remote_mgmt_addr):
            flag = True
        remote_sys_desc = self.pub_lldp.get_remote_sys_desc(tlv_data)
        if attr_obj.remote_sys_desc_uneq_store(remote_sys_desc):
            flag = True
        remote_sys_name = self.pub_lldp.get_remote_sys_name(tlv_data)
        if attr_obj.remote_sys_name_uneq_store(remote_sys_name):
            flag = True
        remote_port = self.pub_lldp.get_remote_port(tlv_data)
        if attr_obj.remote_port_uneq_store(remote_port):
            flag = True
        remote_chassis_id_mac = self.pub_lldp.\
            get_remote_chassis_id_mac(tlv_data)
        if attr_obj.remote_chassis_id_mac_uneq_store(remote_chassis_id_mac):
            flag = True
        remote_port_id_mac = self.pub_lldp.get_remote_port_id_mac(tlv_data)
        if attr_obj.remote_port_id_mac_uneq_store(remote_port_id_mac):
            flag = True
        return flag

    def cfg_lldp_interface(self, protocol_interface, phy_interface=None):
        """Cfg LLDP on interface and create object. """
        if phy_interface is None:
            phy_interface = protocol_interface
        self.create_attr_obj(protocol_interface, phy_interface)
        ret = self.pub_lldp.enable_lldp(protocol_interface)
        attr_obj = self.get_attr_obj(protocol_interface)
        attr_obj.update_lldp_status(ret)

    def cfg_lldp_interface_list(self, intf_list):
        """This routine configures LLDP on the given interfaces list. """
        for intf in intf_list:
            self.cfg_lldp_interface(intf)

    def periodic_discovery_task(self):
        """Periodic task that checks the interface TLV attributes. """
        try:
            self._periodic_task_int()
        except Exception as exc:
            LOG.error("Exception caught in periodic discovery task %s",
                      str(exc))

    def _check_bond_interface_change(self, phy_interface, attr_obj):
        """Check if there's any change in bond interface.

        First check if the interface passed itself is a bond-interface and then
        retrieve the member list and compare.
        Next, check if the interface passed is a part of the bond interface and
        then retrieve the member list and compare.
        """
        bond_phy = sys_utils.get_bond_intf(phy_interface)
        if sys_utils.is_intf_bond(phy_interface):
            bond_intf = phy_interface
        else:
            bond_intf = bond_phy
        # This can be an addition or removal of the interface to a bond.
        bond_intf_change = attr_obj.cmp_update_bond_intf(bond_intf)
        return bond_intf_change

    def _periodic_task_int(self):
        """Internal periodic discovery task routine to check TLV attributes.

        This routine retrieves the LLDP TLC's on all its configured interfaces.
        If the retrieved TLC is different than the stored TLV, it invokes the
        callback.
        """
        for intf in self.intf_list:
            attr_obj = self.get_attr_obj(intf)
            status = attr_obj.get_lldp_status()
            if not status:
                ret = self.pub_lldp.enable_lldp(intf)
                attr_obj.update_lldp_status(ret)
                continue
            bond_intf_change = self._check_bond_interface_change(
                attr_obj.get_phy_interface(), attr_obj)
            tlv_data = self.pub_lldp.get_lldp_tlv(intf)
            # This should take care of storing the information of interest
            if self.cmp_store_tlv_params(intf, tlv_data) or (
                attr_obj.get_db_retry_status() or bond_intf_change or (
                    attr_obj.get_topo_disc_send_cnt() > (
                    constants.TOPO_DISC_SEND_THRESHOLD))):
                # Passing the interface attribute object to CB
                ret = self.cb(intf, attr_obj)
                status = not ret
                attr_obj.store_db_retry_status(status)
                attr_obj.reset_topo_disc_send_cnt()
            else:
                attr_obj.incr_topo_disc_send_cnt()



# Copyright (c) 2013-2016 Cisco Systems, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections

from oslo_config import cfg

from networking_cisco.ml2_drivers.nexus import (
    constants as const)
from networking_cisco.ml2_drivers.nexus import exceptions
from networking_cisco.ml2_drivers.nexus import nexus_db_v2
from networking_cisco.tests.unit.ml2_drivers.nexus import (
    test_cisco_nexus_base)

RP_HOST_NAME_1 = 'UniquePort'
RP_HOST_NAME_2 = 'DuplicateVlan'
RP_HOST_NAME_3 = 'DuplicatePort'
RP_HOST_NAME_DUAL = 'testdualhost'
MAX_REPLAY_COUNT = 4


class TestCiscoNexusReplayResults(
    test_cisco_nexus_base.TestCiscoNexusBaseResults):

    """Unit tests driver results for Cisco ML2 Nexus."""

    test_results = {
        'driver_result_unique_init': (
            [test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('ethernet', '1\/10', 'None')]),

        'driver_result_unique_add1': (
            [test_cisco_nexus_base.RESULT_ADD_VLAN.format(267),
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('ethernet', '1\/10', 267)]),

        'driver_result_unique_add2': (
            [test_cisco_nexus_base.RESULT_ADD_VLAN.format(265),
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('ethernet', '1\/10', 265)]),

        'driver_result_unique_del1': (
            [test_cisco_nexus_base.RESULT_DEL_INTERFACE.
                format('ethernet', '1\/10', 265),
            test_cisco_nexus_base.RESULT_DEL_VLAN.format(265)]),

        'driver_result_unique_del2': (
            [test_cisco_nexus_base.RESULT_DEL_INTERFACE.
                format('ethernet', '1\/10', 267),
            test_cisco_nexus_base.RESULT_DEL_VLAN.format(267)]),

        'driver_result_unique_2vlan_replay': (
            [test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('ethernet', '1\/10', '265,267'),
            test_cisco_nexus_base.RESULT_ADD_VLAN.format('265,267')]),

        'dupl_vlan_result1_add': (
            [test_cisco_nexus_base.RESULT_ADD_VLAN.format(267),
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('ethernet', '1\/10', 267),
            test_cisco_nexus_base.RESULT_ADD_VLAN.format(267),
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('ethernet', '1\/20', 267)]),

        'dupl_vlan_result2_add': (
            [test_cisco_nexus_base.RESULT_ADD_VLAN.format(267),
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('ethernet', '1\/10', 267),
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('ethernet', '1\/20', 267)]),

        'dupl_vlan_result2_del': (
            [test_cisco_nexus_base.RESULT_DEL_INTERFACE.
                format('ethernet', '1\/10', 267),
            test_cisco_nexus_base.RESULT_DEL_VLAN.format(267),
            test_cisco_nexus_base.RESULT_DEL_INTERFACE.
                format('ethernet', '1\/20', 267)]),

        'dupl_vlan_result_replay': (
            [test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('ethernet', '1\/10', 267),
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('ethernet', '1\/20', 267),
            test_cisco_nexus_base.RESULT_ADD_VLAN.format(267)]),

        'dupl_port_result_replay': (
            [test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('ethernet', '1\/10', 267),
            test_cisco_nexus_base.RESULT_ADD_VLAN.format(267)]),

        'switch_up_result_add': (
            [test_cisco_nexus_base.RESULT_ADD_VLAN.format(269),
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('ethernet', '1\/3', 269)]),

        'switch_up_result_del': (
            [test_cisco_nexus_base.RESULT_DEL_INTERFACE.
                format('ethernet', '1\/3', 269),
            test_cisco_nexus_base.RESULT_DEL_VLAN.format(269)]),

        'switch_restore_result_add': (
            [test_cisco_nexus_base.RESULT_ADD_VLAN.format(269),
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('ethernet', '1\/3', 269)]),

        'switch_restore_result_replay': (
            [test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('ethernet', '1\/2', 269),
            test_cisco_nexus_base.RESULT_ADD_VLAN.format(269)]),

        'switch_restore_result_del': (
            [test_cisco_nexus_base.RESULT_DEL_INTERFACE.
                format('ethernet', '1\/3', 269),
            test_cisco_nexus_base.RESULT_DEL_VLAN.format(269),
            test_cisco_nexus_base.RESULT_DEL_INTERFACE.
                format('ethernet', '1\/2', 269),
            test_cisco_nexus_base.RESULT_DEL_VLAN.format(269)]),

    }


class TestCiscoNexusReplay(test_cisco_nexus_base.TestCiscoNexusReplayBase):
    """Unit tests for Replay of Cisco ML2 Nexus data."""
    test_configs = {
        'test_replay_unique1':
            test_cisco_nexus_base.TestCiscoNexusBase.TestConfigObj(
                test_cisco_nexus_base.NEXUS_IP_ADDRESS_1,
                RP_HOST_NAME_1,
                test_cisco_nexus_base.NEXUS_PORT_1,
                test_cisco_nexus_base.INSTANCE_1,
                test_cisco_nexus_base.VLAN_ID_1,
                test_cisco_nexus_base.NO_VXLAN_ID,
                None,
                test_cisco_nexus_base.DEVICE_OWNER_COMPUTE,
                {},
                None,
                test_cisco_nexus_base.NORMAL_VNIC),
        'test_replay_unique2':
            test_cisco_nexus_base.TestCiscoNexusBase.TestConfigObj(
                test_cisco_nexus_base.NEXUS_IP_ADDRESS_1,
                RP_HOST_NAME_1,
                test_cisco_nexus_base.NEXUS_PORT_1,
                test_cisco_nexus_base.INSTANCE_2,
                test_cisco_nexus_base.VLAN_ID_2,
                test_cisco_nexus_base.NO_VXLAN_ID,
                None,
                test_cisco_nexus_base.DEVICE_OWNER_COMPUTE,
                {},
                None,
                test_cisco_nexus_base.NORMAL_VNIC),
        'test_replay_duplvlan1':
            test_cisco_nexus_base.TestCiscoNexusBase.TestConfigObj(
                test_cisco_nexus_base.NEXUS_IP_ADDRESS_2,
                RP_HOST_NAME_2,
                test_cisco_nexus_base.NEXUS_PORT_1,
                test_cisco_nexus_base.INSTANCE_1,
                test_cisco_nexus_base.VLAN_ID_1,
                test_cisco_nexus_base.NO_VXLAN_ID,
                None,
                test_cisco_nexus_base.DEVICE_OWNER_COMPUTE,
                {},
                None,
                test_cisco_nexus_base.NORMAL_VNIC),
        'test_replay_duplvlan2':
            test_cisco_nexus_base.TestCiscoNexusBase.TestConfigObj(
                test_cisco_nexus_base.NEXUS_IP_ADDRESS_2,
                RP_HOST_NAME_2,
                test_cisco_nexus_base.NEXUS_PORT_2,
                test_cisco_nexus_base.INSTANCE_2,
                test_cisco_nexus_base.VLAN_ID_1,
                test_cisco_nexus_base.NO_VXLAN_ID,
                None,
                test_cisco_nexus_base.DEVICE_OWNER_COMPUTE,
                {},
                None,
                test_cisco_nexus_base.NORMAL_VNIC),
        'test_replay_duplport1':
            test_cisco_nexus_base.TestCiscoNexusBase.TestConfigObj(
                test_cisco_nexus_base.NEXUS_IP_ADDRESS_3,
                RP_HOST_NAME_3,
                test_cisco_nexus_base.NEXUS_PORT_1,
                test_cisco_nexus_base.INSTANCE_1,
                test_cisco_nexus_base.VLAN_ID_1,
                test_cisco_nexus_base.NO_VXLAN_ID,
                None,
                test_cisco_nexus_base.DEVICE_OWNER_COMPUTE,
                {},
                None,
                test_cisco_nexus_base.NORMAL_VNIC),
        'test_replay_duplport2':
            test_cisco_nexus_base.TestCiscoNexusBase.TestConfigObj(
                test_cisco_nexus_base.NEXUS_IP_ADDRESS_3,
                RP_HOST_NAME_3,
                test_cisco_nexus_base.NEXUS_PORT_1,
                test_cisco_nexus_base.INSTANCE_2,
                test_cisco_nexus_base.VLAN_ID_1,
                test_cisco_nexus_base.NO_VXLAN_ID,
                None,
                test_cisco_nexus_base.DEVICE_OWNER_COMPUTE,
                {},
                None,
                test_cisco_nexus_base.NORMAL_VNIC),
        'test_replay_dual':
            test_cisco_nexus_base.TestCiscoNexusBase.TestConfigObj(
                test_cisco_nexus_base.NEXUS_IP_ADDRESS_DUAL,
                RP_HOST_NAME_DUAL,
                test_cisco_nexus_base.NEXUS_DUAL1,
                test_cisco_nexus_base.INSTANCE_DUAL,
                test_cisco_nexus_base.VLAN_ID_DUAL,
                test_cisco_nexus_base.NO_VXLAN_ID,
                None,
                test_cisco_nexus_base.DEVICE_OWNER_COMPUTE,
                {},
                None,
                test_cisco_nexus_base.NORMAL_VNIC),
        'test_replay_dual2':
            test_cisco_nexus_base.TestCiscoNexusBase.TestConfigObj(
                test_cisco_nexus_base.NEXUS_IP_ADDRESS_DUAL2,
                RP_HOST_NAME_DUAL,
                test_cisco_nexus_base.NEXUS_DUAL2,
                test_cisco_nexus_base.INSTANCE_DUAL,
                test_cisco_nexus_base.VLAN_ID_DUAL,
                test_cisco_nexus_base.NO_VXLAN_ID,
                None,
                test_cisco_nexus_base.DEVICE_OWNER_COMPUTE,
                {},
                None,
                test_cisco_nexus_base.NORMAL_VNIC),
        'test_replay_vxlan_unique1':
            test_cisco_nexus_base.TestCiscoNexusBase.TestConfigObj(
                test_cisco_nexus_base.NEXUS_IP_ADDRESS_1,
                RP_HOST_NAME_1,
                test_cisco_nexus_base.NEXUS_PORT_1,
                test_cisco_nexus_base.INSTANCE_1,
                test_cisco_nexus_base.VLAN_ID_1,
                test_cisco_nexus_base.VXLAN_ID,
                test_cisco_nexus_base.MCAST_GROUP,
                test_cisco_nexus_base.DEVICE_OWNER_COMPUTE,
                {},
                None,
                test_cisco_nexus_base.NORMAL_VNIC),
    }
    test_configs = collections.OrderedDict(sorted(test_configs.items()))

    def setUp(self):
        """Sets up mock ncclient, and switch and credentials dictionaries."""

        cfg.CONF.set_override('nexus_driver', 'ncclient', 'ml2_cisco')
        cfg.CONF.set_override('never_cache_ssh_connection', False, 'ml2_cisco')
        super(TestCiscoNexusReplay, self).setUp()
        self.results = TestCiscoNexusReplayResults()

    def test_replay_unique_ports(self):
        """Provides replay data and result data for unique ports. """

        first_add = {
            'driver_results': self.results.get_test_results(
                'driver_result_unique_add1'),
            'nbr_db_entries': 1}
        second_add = {
            'driver_results': self.results.get_test_results(
                'driver_result_unique_add2'),
            'nbr_db_entries': 2}
        first_del = {
            'driver_results': self.results.get_test_results(
                'driver_result_unique_del1'),
            'nbr_db_entries': 1}
        second_del = {
            'driver_results': self.results.get_test_results(
                'driver_result_unique_del2'),
            'nbr_db_entries': 0}

        self._process_replay(
            'test_replay_unique1',
            'test_replay_unique2',
            self.results.get_test_results(
                'driver_result_unique_init'),
            first_add,
            second_add,
            self.results.get_test_results(
                'driver_result_unique_2vlan_replay'),
            first_del,
            second_del)

    def test_replay_duplicate_vlan(self):
        """Provides replay data and result data for duplicate vlans. """

        first_add = {
            'driver_results': self.results.get_test_results(
                'dupl_vlan_result1_add'),
            'nbr_db_entries': 2}

        # TODO(caboucha)
        # 'driver_result': [], until the correct fix for
        # the following issue is resolved.
        # https://review.openstack.org/#/c/241216/
        second_add = {
            'driver_results': self.results.get_test_results(
                'dupl_vlan_result2_add'),
            'nbr_db_entries': 4}

        first_del = {'driver_results': [],
                     'nbr_db_entries': 2}

        second_del = {
            'driver_results': self.results.get_test_results(
                'dupl_vlan_result2_del'),
            'nbr_db_entries': 0}

        self._process_replay('test_replay_duplvlan1',
                             'test_replay_duplvlan2',
                             [],
                             first_add, second_add,
                             self.results.get_test_results(
                                 'dupl_vlan_result_replay'),
                             first_del, second_del)

    def test_replay_duplicate_ports(self):
        """Provides replay data and result data for duplicate ports. """

        first_add = {
            'driver_results': self.results.get_test_results(
                'driver_result_unique_add1'),
            'nbr_db_entries': 1}

        # TODO(caboucha)
        # 'driver_result': [], until the correct fix for
        # the following issue is resolved.
        # https://review.openstack.org/#/c/241216/
        second_add = {
            'driver_results': self.results.get_test_results(
                'driver_result_unique_add1'),
            'nbr_db_entries': 2}
        first_del = {'driver_results': [],
                     'nbr_db_entries': 1}

        second_del = {
            'driver_results': self.results.get_test_results(
                'driver_result_unique_del2'),
            'nbr_db_entries': 0}

        self._process_replay('test_replay_duplport1',
                             'test_replay_duplport2',
                             [],
                             first_add, second_add,
                             self.results.get_test_results(
                                 'dupl_port_result_replay'),
                             first_del, second_del)

    def test_replay_enable_vxlan_feature_failure(self):
        """Verifies exception during enable VXLAN feature driver. """

        # Set configuration variable to add/delete the VXLAN global nexus
        # switch values.
        cfg.CONF.set_override('vxlan_global_config', True, 'ml2_cisco')
        self._create_port_failure(
            'connect.return_value.edit_config.side_effect',
            'feature nv overlay vn-segment-vlan-based',
            'test_replay_vxlan_unique1',
            __name__)

    def test_replay_disable_vxlan_feature_failure(self):
        """Verifies exception during disable VXLAN feature driver. """

        # Set configuration variable to add/delete the VXLAN global nexus
        # switch values.
        cfg.CONF.set_override('vxlan_global_config', True, 'ml2_cisco')
        self._delete_port_failure(
            'connect.return_value.edit_config.side_effect',
            'no feature nv overlay vn-segment-vlan-based',
            'test_replay_vxlan_unique1',
            __name__)

    def test_replay_create_nve_member_failure(self):
        """Verifies exception during create nve member driver. """

        self._create_port_failure(
            'connect.return_value.edit_config.side_effect',
            'member vni mcast-group',
            'test_replay_vxlan_unique1',
            __name__)

    def test_replay_delete_nve_member_failure(self):
        """Verifies exception during delete nve member driver. """

        self._delete_port_failure(
            'connect.return_value.edit_config.side_effect',
            'no member vni',
            'test_replay_vxlan_unique1',
            __name__)

    def test_replay_create_vlan_failure(self):
        """Verifies exception during edit vlan create driver. """

        self._create_port_failure(
            'connect.return_value.edit_config.side_effect',
            'vlan-id-create-delete',
            'test_replay_unique1',
            __name__)

    def test_replay_delete_vlan_failure(self):
        """Verifies exception during edit vlan delete driver. """

        self._delete_port_failure(
            'connect.return_value.edit_config.side_effect',
            'vlan-id-create-delete no vlan 267',
            'test_replay_unique1',
            __name__)

    def test_replay_create_trunk_failure(self):
        """Verifies exception during create trunk interface driver. """

        self._create_port_failure(
            'connect.return_value.edit_config.side_effect',
            'switchport trunk allowed vlan_id 267',
            'test_replay_unique1',
            __name__)

    def test_replay_delete_trunk_failure(self):
        """Verifies exception during delete trunk interface driver. """

        self._delete_port_failure(
            'connect.return_value.edit_config.side_effect',
            'switchport trunk allowed remove vlan 267',
            'test_replay_unique1',
            __name__)

    def test_replay_new_port_success_if_one_switch_up(self):
        """Verifies create port successful if one multi-switch up."""

        # Make sure port is not rejected when there are multiple
        # switches and only one is active.
        port_cfg1 = self.test_configs['test_replay_dual']
        self._cisco_mech_driver.set_switch_ip_and_active_state(
            port_cfg1.nexus_ip_addr, const.SWITCH_ACTIVE)
        port_cfg2 = self.test_configs['test_replay_dual2']
        self._cisco_mech_driver.set_switch_ip_and_active_state(
            port_cfg2.nexus_ip_addr, const.SWITCH_INACTIVE)

        # Set-up successful creation of port vlan config
        self._basic_create_verify_port_vlan('test_replay_dual',
            self.results.get_test_results(
                'switch_up_result_add'),
            nbr_of_bindings=1)

        # Even though 2nd entry is inactive, there should be
        # a data base entry configured for it.
        # 2 = One entry for port the other for reserved binding
        self.assertEqual(
            2, len(nexus_db_v2.get_nexusport_switch_bindings(
                   port_cfg2.nexus_ip_addr)))

        # Clean-up the port entry
        self._basic_delete_verify_port_vlan('test_replay_dual',
            self.results.get_test_results(
                'switch_up_result_del'),
            nbr_of_bindings=0)

    def test_replay_port_success_if_one_switch_restored(self):
        """Verifies port restored after one of multi-switch restored."""

        # Make sure port is not rejected when there are multiple
        # switches and one is active.  Then proceed to bring-up
        # the other switch and it gets configured successfully.
        # Then remove all.
        port_cfg1 = self.test_configs['test_replay_dual']
        self._cisco_mech_driver.set_switch_ip_and_active_state(
            port_cfg1.nexus_ip_addr, const.SWITCH_ACTIVE)
        port_cfg2 = self.test_configs['test_replay_dual2']
        self._cisco_mech_driver.set_switch_ip_and_active_state(
            port_cfg2.nexus_ip_addr, const.SWITCH_INACTIVE)

        # Set-up successful creation of port vlan config
        self._basic_create_verify_port_vlan('test_replay_dual',
            self.results.get_test_results(
                'switch_restore_result_add'),
            nbr_of_bindings=1)

        # Even though 2nd entry is inactive, there should be
        # a data base entry configured for it.
        # 2 = One entry for port the other for reserved binding
        self.assertEqual(
            2, len(nexus_db_v2.get_nexusport_switch_bindings(
                   port_cfg2.nexus_ip_addr)))

        # Restore port data for that switch
        self._cfg_monitor.check_connections()
        self._verify_results(
            self.results.get_test_results(
                'switch_restore_result_replay'))

        # Clear mock_call history.
        self.mock_ncclient.reset_mock()

        # Clean-up the port entries
        self._basic_delete_verify_port_vlan('test_replay_dual',
            self.results.get_test_results(
                'switch_restore_result_del'),
            nbr_of_bindings=0)

    def test_replay_create_fails_if_single_switch_down(self):
        """Verifies port create fails if switch down."""

        # Make sure create ethernet config fails when the
        # switch state is inactive.
        port_cfg = self.test_configs['test_replay_unique1']
        self._cisco_mech_driver.set_switch_ip_and_active_state(
            port_cfg.nexus_ip_addr, const.SWITCH_INACTIVE)

        port_context = self._generate_port_context(port_cfg)
        self.assertRaises(
                exceptions.NexusConnectFailed,
                self._cisco_mech_driver.create_port_postcommit,
                port_context)

    def test_replay_update_fails_if_single_switch_down(self):
        """Verifies port update fails if switch down."""

        # Make sure update ethernet config fails when the
        # switch state is inactive.
        port_cfg = self.test_configs['test_replay_unique1']
        self._cisco_mech_driver.set_switch_ip_and_active_state(
            port_cfg.nexus_ip_addr, const.SWITCH_INACTIVE)

        port_context = self._generate_port_context(port_cfg)
        self.assertRaises(
                exceptions.NexusConnectFailed,
                self._cisco_mech_driver.update_port_postcommit,
                port_context)

    def test_replay_delete_success_if_switch_down(self):
        """Verifies port delete success if switch down."""

        # Make sure delete config successful even when the
        # switch state is inactive.
        port_cfg = self.test_configs['test_replay_unique1']
        self._cisco_mech_driver.set_switch_ip_and_active_state(
            port_cfg.nexus_ip_addr, const.SWITCH_ACTIVE)

        self._basic_create_verify_port_vlan('test_replay_unique1',
            self.results.get_test_results(
                'driver_result_unique_add1'))

        # Make switch inactive before delete
        self._cisco_mech_driver.set_switch_ip_and_active_state(
            port_cfg.nexus_ip_addr, const.SWITCH_INACTIVE)

        # Clean-up the port entry
        self._basic_delete_verify_port_vlan('test_replay_unique1',
            [], nbr_of_bindings=0)

    def test_replay_get_nexus_type_failure_two_switches(self):
        """Verifies exception during ncclient get inventory. """

        # There are two switches, one active and the other inactive.
        # Make sure 'get_nexus_type' fails so create_port_postcommit()
        # will return an exception.  'get_nexus_type' is used as
        # as ping so even if the switch is marked active then double
        # check it is indeed still active.  If not and thre are no
        # other active switches, then raise exception.
        port_cfg1 = self.test_configs['test_replay_dual']
        self._cisco_mech_driver.set_switch_ip_and_active_state(
            port_cfg1.nexus_ip_addr, const.SWITCH_ACTIVE)
        port_cfg2 = self.test_configs['test_replay_dual2']
        self._cisco_mech_driver.set_switch_ip_and_active_state(
            port_cfg2.nexus_ip_addr, const.SWITCH_INACTIVE)

        # Set-up so get_nexus_type driver fails on active switch
        config = {'connect.return_value.get.side_effect':
            self._config_side_effects_on_count('show inventory',
            Exception(__name__))}
        self.mock_ncclient.configure_mock(**config)

        port_context = self._generate_port_context(port_cfg1)
        self.assertRaises(
                exceptions.NexusConnectFailed,
                self._cisco_mech_driver.create_port_postcommit,
                port_context)

    def test_replay_get_nexus_type_failure(self):
        """Verifies exception during get nexus_type while replaying. """

        # Set switch state to False so replay config will start.
        # This should not affect user configuration.
        port_cfg = self.test_configs['test_replay_unique1']
        self._cisco_mech_driver.set_switch_ip_and_active_state(
            port_cfg.nexus_ip_addr, const.SWITCH_ACTIVE)

        # Set-up successful creation of port vlan config
        self._basic_create_verify_port_vlan('test_replay_unique1',
            self.results.get_test_results(
                'driver_result_unique_add1'))

        # Set-up so get_nexus_type driver fails
        config = {'connect.return_value.get.side_effect':
            self._config_side_effects_on_count('show inventory',
            Exception(__name__))}
        self.mock_ncclient.configure_mock(**config)

        self._cisco_mech_driver.set_switch_ip_and_active_state(
            port_cfg.nexus_ip_addr, const.SWITCH_INACTIVE)

        # Perform replay which should not send back exception
        # but merely quit
        self._cfg_monitor.check_connections()

        # Since get of nexus_type failed, there should be
        # no attempt to configure anything.
        self._verify_results([])

        # Clean-up the port entry
        self._basic_delete_verify_port_vlan('test_replay_unique1',
            [])

    def test_replay_create_vlan_failure_during_replay(self):
        """Verifies exception during create vlan while replaying. """

        vlan267 = test_cisco_nexus_base.RESULT_ADD_VLAN.format(267)
        driver_result1 = [vlan267] * 2

        # Set switch state to False so replay config will start.
        # This should not affect user configuration.
        port_cfg = self.test_configs['test_replay_unique1']
        self._cisco_mech_driver.set_switch_ip_and_active_state(
            port_cfg.nexus_ip_addr, const.SWITCH_ACTIVE)

        # Set-up successful creation of port vlan config
        self._basic_create_verify_port_vlan('test_replay_unique1',
            self.results.get_test_results(
                'driver_result_unique_add1'))

        # Set-up exception during create_vlan
        config = {'connect.return_value.edit_config.side_effect':
                  self._config_side_effects_on_count(
                      'vlan-id-create-delete',
                      Exception(__name__ + '1'))}
        self.mock_ncclient.configure_mock(**config)

        self._cisco_mech_driver.set_switch_ip_and_active_state(
            port_cfg.nexus_ip_addr, const.SWITCH_INACTIVE)

        # Perform replay which should not send back exception
        # but merely quit
        self._cfg_monitor.check_connections()

        # Verify that switch is put back into INACTIVE state
        self.assertEqual(
            const.SWITCH_INACTIVE,
            self._cisco_mech_driver.get_switch_ip_and_active_state(
                port_cfg.nexus_ip_addr))

        # The edit of create_vlan failed, but there will
        # be 2 create vlan attempts in mock call history.
        result_replay = (
            [test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('ethernet', '1\/10', 267)] + driver_result1)

        self._verify_results(result_replay)

        # Clear the edit driver exception for next test.
        config = {'connect.return_value.edit_config.side_effect':
                  None}
        self.mock_ncclient.configure_mock(**config)

        # Perform replay which should not send back exception
        # but merely quit
        self._cfg_monitor.check_connections()

        # Verify that switch is in ACTIVE state
        self.assertEqual(
            const.SWITCH_ACTIVE,
            self._cisco_mech_driver.get_switch_ip_and_active_state(
                port_cfg.nexus_ip_addr))

        # Clear mock_call history.
        self.mock_ncclient.reset_mock()

        # Clean-up the port entry
        self._basic_delete_verify_port_vlan('test_replay_unique1',
            self.results.get_test_results(
                'driver_result_unique_del2'))

    def test_replay_vlan_batch_failure_during_replay(self):
        """Verifies handling of batch vlan during replay."""

        tmp_cfg = self.TestConfigObj(
            test_cisco_nexus_base.NEXUS_IP_ADDRESS_1,
            RP_HOST_NAME_1,
            test_cisco_nexus_base.NEXUS_PORT_1,
            test_cisco_nexus_base.INSTANCE_1,
            test_cisco_nexus_base.VLAN_ID_1,
            test_cisco_nexus_base.NO_VXLAN_ID,
            None,
            test_cisco_nexus_base.DEVICE_OWNER_COMPUTE,
            {},
            None,
            test_cisco_nexus_base.NORMAL_VNIC)
        self._cisco_mech_driver.set_switch_ip_and_active_state(
            tmp_cfg.nexus_ip_addr, const.SWITCH_ACTIVE)

        # Create a batch of port entries with unique vlans
        num_vlans = const.CREATE_VLAN_BATCH + 10
        for x in range(num_vlans):
            instance_id = test_cisco_nexus_base.INSTANCE_1 + '-' + str(x)
            new_cfg = tmp_cfg._replace(
                          vlan_id=test_cisco_nexus_base.VLAN_ID_1 + x,
                          instance_id=instance_id)
            self._create_port(new_cfg)

        # Put it back to inactive state
        self._cisco_mech_driver.set_switch_ip_and_active_state(
            tmp_cfg.nexus_ip_addr, const.SWITCH_INACTIVE)

        self._cfg_monitor.check_connections()

        # Verify the switch is in restore stage 2 state
        self.assertEqual(
            const.SWITCH_RESTORE_S2,
            self._cisco_mech_driver.get_switch_ip_and_active_state(
                tmp_cfg.nexus_ip_addr))

        config = {'connect.return_value.edit_config.side_effect':
                  self._config_side_effects_on_count(
                      'vlan-id-create-delete',
                      Exception(__name__ + '1'))}
        self.mock_ncclient.configure_mock(**config)

        # Call check_connections() again  to attempt to send
        # last batch of 10 which should fail
        self._cfg_monitor.check_connections()
        # Verify the switch is back in INACTIVE state
        self.assertEqual(
            const.SWITCH_INACTIVE,
            self._cisco_mech_driver.get_switch_ip_and_active_state(
                tmp_cfg.nexus_ip_addr))

        # Clear mock_call history.
        self.mock_ncclient.reset_mock()

        # Clear the edit driver exception for next test.
        config = {'connect.return_value.edit_config.side_effect':
                  None}
        self.mock_ncclient.configure_mock(**config)

        # Call check_connections() again  to restart restore
        self._cfg_monitor.check_connections()

        # Verify the switch is in restore stage 2 state
        self.assertEqual(
            const.SWITCH_RESTORE_S2,
            self._cisco_mech_driver.get_switch_ip_and_active_state(
                tmp_cfg.nexus_ip_addr))

        # Call check_connections() to successfully send
        # last batch of 10 which should fail
        self._cfg_monitor.check_connections()

        # Verify the switch is in restore stage 2 state
        self.assertEqual(
            const.SWITCH_ACTIVE,
            self._cisco_mech_driver.get_switch_ip_and_active_state(
                tmp_cfg.nexus_ip_addr))

    def test_replay_no_retry_failure_handling(self):
        """Tests to check replay 'no retry' failure handling.

        1) Verify config_failure is incremented upon failure during
        replay config and verify create_vlan transactions are seen.
        2) Verify contact_failure is incremented upon failure during
        get_nexus_type transaction.
        3) Verify receipt of new transaction does not reset
        failure statistics.
        4) Verify config&contact_failure is reset when replay is
        successful.
        """

        # Due to 2 retries in driver to deal with stale ncclient
        # handle, the results are doubled.
        vlan267 = '<vlan-id-create-delete\>\s+\<__XML__PARAM_value\>267'
        driver_result2 = ([test_cisco_nexus_base.RESULT_ADD_INTERFACE.
            format('ethernet', '1\/10', 267)] + [vlan267] * 2) * 4

        config_replay = MAX_REPLAY_COUNT
        port_cfg = self.test_configs['test_replay_unique1']
        self._cisco_mech_driver.set_switch_ip_and_active_state(
            port_cfg.nexus_ip_addr, const.SWITCH_ACTIVE)

        # Set-up successful creation of port vlan config
        self._basic_create_verify_port_vlan(
            'test_replay_unique1',
            self.results.get_test_results(
                'driver_result_unique_add1'))

        # Test 1:
        # Set the edit create vlan driver exception
        # Perform replay MAX_REPLAY_COUNT times
        # This should not roll-up an exception but merely quit
        # and increment FAIL_CONFIG statistics

        config = {'connect.return_value.edit_config.side_effect':
                  self._config_side_effects_on_count(
                      'vlan-id-create-delete',
                      Exception(__name__ + '1'))}
        self.mock_ncclient.configure_mock(**config)

        self._cisco_mech_driver.set_switch_ip_and_active_state(
            port_cfg.nexus_ip_addr, const.SWITCH_INACTIVE)
        for i in range(config_replay):
            self._cfg_monitor.check_connections()

        # Verify FAIL_CONFIG reached(MAX_REPLAY_COUNT) and there
        # were only MAX_REPLAY_COUNT+1 attempts to send create_vlan.
        # first is from test_replay_create_vlan_failure()
        # and MAX_REPLAY_COUNT from check_connections()
        self.assertEqual(
            config_replay,
            self._cisco_mech_driver.get_switch_replay_failure(
                const.FAIL_CONFIG,
                test_cisco_nexus_base.NEXUS_IP_ADDRESS_1))
        self._verify_results(driver_result2)

        # Verify there exists a single port binding
        # plus 1 for reserved switch entry
        self.assertEqual(
            2, len(nexus_db_v2.get_nexusport_switch_bindings(
                   test_cisco_nexus_base.NEXUS_IP_ADDRESS_1)))

        # Clear mock_call history.
        self.mock_ncclient.reset_mock()

        # Clear the edit driver exception for next test.
        config = {'connect.return_value.edit_config.side_effect':
                  None}
        self.mock_ncclient.configure_mock(**config)

        # Test 2)
        # Set it up so get nexus type returns exception.
        # FAIL_CONTACT should increment.

        self._set_nexus_type_failure()

        # Perform replay MAX_REPLAY_COUNT times
        # This should not roll-up an exception but merely quit
        for i in range(config_replay):
            self._cfg_monitor.check_connections()

        # Verify switch FAIL_CONTACT reached (MAX_REPLAY_COUNT)
        # and there were no attempts to send create_vlan.
        self.assertEqual(
            config_replay,
            self._cisco_mech_driver.get_switch_replay_failure(
                const.FAIL_CONFIG,
                test_cisco_nexus_base.NEXUS_IP_ADDRESS_1))
        self.assertEqual(
            config_replay,
            self._cisco_mech_driver.get_switch_replay_failure(
                const.FAIL_CONTACT,
                test_cisco_nexus_base.NEXUS_IP_ADDRESS_1))
        self._verify_results([])

        # Test 3)
        # Verify delete transaction doesn't affect failure stats.
        self._basic_delete_verify_port_vlan('test_replay_unique1',
            [])

        # Verify failure stats is not reset
        self.assertEqual(
            config_replay,
            self._cisco_mech_driver.get_switch_replay_failure(
                const.FAIL_CONFIG,
                test_cisco_nexus_base.NEXUS_IP_ADDRESS_1))
        self.assertEqual(
            config_replay,
            self._cisco_mech_driver.get_switch_replay_failure(
                const.FAIL_CONTACT,
                test_cisco_nexus_base.NEXUS_IP_ADDRESS_1))

        # Clear the get nexus type driver exception.
        config = {'connect.return_value.get.side_effect':
                  None}
        self.mock_ncclient.configure_mock(**config)

        # Test 4)
        # Verify config&contact_failure is reset when replay is
        # successful.

        # Perform replay once which will be successful causing
        # failure stats to be reset to 0.
        # Then verify these stats are indeed 0.
        self._cfg_monitor.check_connections()
        self.assertEqual(
            0,
            self._cisco_mech_driver.get_switch_replay_failure(
                const.FAIL_CONFIG,
                test_cisco_nexus_base.NEXUS_IP_ADDRESS_1))
        self.assertEqual(
            0,
            self._cisco_mech_driver.get_switch_replay_failure(
                const.FAIL_CONTACT,
                test_cisco_nexus_base.NEXUS_IP_ADDRESS_1))

        # Verify switch state is now active following successful replay.
        self.assertEqual(
            const.SWITCH_ACTIVE,
            self._cisco_mech_driver.get_switch_ip_and_active_state(
                test_cisco_nexus_base.NEXUS_IP_ADDRESS_1))


class TestCiscoNexusBaremetalReplayResults(
    test_cisco_nexus_base.TestCiscoNexusBaseResults):
    """Unit tests driver results for Cisco ML2 Nexus."""

    test_results = {

        'driver_result_unique_eth_init': (
            [test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('ethernet', '1\/10', 'None')],
            [test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('ethernet', '1\/20', 'None')],
        ),

        'driver_result_unique_eth_add1': (
            [test_cisco_nexus_base.RESULT_ADD_VLAN.format(267),
            (test_cisco_nexus_base.RESULT_ADD_NATIVE_INTERFACE.
                format('ethernet', '1\/10', 267) +
            '[\x00-\x7f]+' +
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('ethernet', '1\/10', 267))]),

        'driver_result_unique_eth_add2': (
            [test_cisco_nexus_base.RESULT_ADD_VLAN.format(265),
            (test_cisco_nexus_base.RESULT_ADD_NATIVE_INTERFACE.
                format('ethernet', '1\/20', 265) +
            '[\x00-\x7f]+' +
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('ethernet', '1\/20', 265))]),

        'driver_result_unique_eth_del1': (
            [(test_cisco_nexus_base.RESULT_DEL_NATIVE_INTERFACE.
                format('ethernet', '1\/20') +
            '[\x00-\x7f]+' +
            test_cisco_nexus_base.RESULT_DEL_INTERFACE.
                format('ethernet', '1\/20', 265)),
            test_cisco_nexus_base.RESULT_DEL_VLAN.format(265)]),

        'driver_result_unique_eth_del2': (
            [(test_cisco_nexus_base.RESULT_DEL_NATIVE_INTERFACE.
                format('ethernet', '1\/10') +
            '[\x00-\x7f]+' +
            test_cisco_nexus_base.RESULT_DEL_INTERFACE.
                format('ethernet', '1\/10', 267)),
            test_cisco_nexus_base.RESULT_DEL_VLAN.format(267)]),

        'driver_result_unique_2if_replay': (
            [(test_cisco_nexus_base.RESULT_ADD_NATIVE_INTERFACE.
            format('ethernet', '1\/10', 267) +
            '[\x00-\x7f]+' +
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('ethernet', '1\/10', 267)),
            (test_cisco_nexus_base.RESULT_ADD_NATIVE_INTERFACE.
            format('ethernet', '1\/20', 265) +
            '[\x00-\x7f]+' +
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('ethernet', '1\/20', 265)),
            test_cisco_nexus_base.RESULT_ADD_VLAN.format('265,267')]),

        'driver_result_unique_eth_add_vm': (
            [test_cisco_nexus_base.RESULT_ADD_VLAN.format(265),
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('ethernet', '1\/10', 265)]),

        'driver_result_unique_eth_del_vm': (
            [test_cisco_nexus_base.RESULT_DEL_INTERFACE.
                format('ethernet', '1\/10', 265),
            test_cisco_nexus_base.RESULT_DEL_VLAN.format(265)]),

        'driver_result_unique_2vlan_replay': (
            [(test_cisco_nexus_base.RESULT_ADD_NATIVE_INTERFACE.
            format('ethernet', '1\/10', 267) +
            '[\x00-\x7f]+' +
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('ethernet', '1\/10', 267)),
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('ethernet', '1\/10', 265),
            test_cisco_nexus_base.RESULT_ADD_VLAN.format('265,267')]),

        'driver_result_unique_vPC_2switch_add1': (
            [test_cisco_nexus_base.RESULT_ADD_VLAN.format(267),
            (test_cisco_nexus_base.RESULT_ADD_NATIVE_INTERFACE.
                format('port-channel', '469', 267) +
            '[\x00-\x7f]+' +
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('port-channel', '469', 267)),
            test_cisco_nexus_base.RESULT_ADD_VLAN.format(267),
            (test_cisco_nexus_base.RESULT_ADD_NATIVE_INTERFACE.
                format('port-channel', '469', 267) +
            '[\x00-\x7f]+' +
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('port-channel', '469', 267))]),

        'driver_result_unique_vPC_2switch_del1': (
            [(test_cisco_nexus_base.RESULT_DEL_NATIVE_INTERFACE.
                format('port-channel', '469') +
            '[\x00-\x7f]+' +
            test_cisco_nexus_base.RESULT_DEL_INTERFACE.
                format('port-channel', '469', 267)),
            test_cisco_nexus_base.RESULT_DEL_VLAN.format(267),
            (test_cisco_nexus_base.RESULT_DEL_NATIVE_INTERFACE.
                format('port-channel', '469') +
            '[\x00-\x7f]+' +
            test_cisco_nexus_base.RESULT_DEL_INTERFACE.
                format('port-channel', '469', 267)),
            test_cisco_nexus_base.RESULT_DEL_VLAN.format(267)]),

        'driver_result_unique_vPC_2if_replay': (
            [(test_cisco_nexus_base.RESULT_ADD_NATIVE_INTERFACE.
                format('port-channel', '469', 267) +
            '[\x00-\x7f]+' +
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('port-channel', '469', 267)),
            test_cisco_nexus_base.RESULT_ADD_VLAN.format(267),
            (test_cisco_nexus_base.RESULT_ADD_NATIVE_INTERFACE.
                format('port-channel', '469', 267) +
            '[\x00-\x7f]+' +
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('port-channel', '469', 267)),
            test_cisco_nexus_base.RESULT_ADD_VLAN.format(267)]),

        'driver_result_unique_vPC470_add1': (
            [test_cisco_nexus_base.RESULT_ADD_VLAN.format(267),
            (test_cisco_nexus_base.RESULT_ADD_NATIVE_INTERFACE.
                format('port-channel', '470', 267) +
            '[\x00-\x7f]+' +
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('port-channel', '470', 267)),
            test_cisco_nexus_base.RESULT_ADD_VLAN.format(267),
            (test_cisco_nexus_base.RESULT_ADD_NATIVE_INTERFACE.
                format('port-channel', '470', 267) +
            '[\x00-\x7f]+' +
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('port-channel', '470', 267))]),

        'driver_result_unique_vPC470_add2': (
            [test_cisco_nexus_base.RESULT_ADD_VLAN.format(265),
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('port-channel', '470', 265),
            test_cisco_nexus_base.RESULT_ADD_VLAN.format(265),
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('port-channel', '470', 265)]),

        'driver_result_unique_vPC470_del1': (
            [test_cisco_nexus_base.RESULT_DEL_INTERFACE.
                format('port-channel', '470', 265),
            test_cisco_nexus_base.RESULT_DEL_VLAN.format(265),
            test_cisco_nexus_base.RESULT_DEL_INTERFACE.
                format('port-channel', '470', 265),
            test_cisco_nexus_base.RESULT_DEL_VLAN.format(265)]),

        'driver_result_unique_vPC470_del2': (
            [(test_cisco_nexus_base.RESULT_DEL_NATIVE_INTERFACE.
                format('port-channel', '470') +
            '[\x00-\x7f]+' +
            test_cisco_nexus_base.RESULT_DEL_INTERFACE.
                format('port-channel', '470', 267)),
            test_cisco_nexus_base.RESULT_DEL_VLAN.format(267),
            (test_cisco_nexus_base.RESULT_DEL_NATIVE_INTERFACE.
                format('port-channel', '470') +
            '[\x00-\x7f]+' +
            test_cisco_nexus_base.RESULT_DEL_INTERFACE.
                format('port-channel', '470', 267)),
            test_cisco_nexus_base.RESULT_DEL_VLAN.format(267)]),

        'driver_result_unique_vPC470_2vlan_replay': (
            [(test_cisco_nexus_base.RESULT_ADD_NATIVE_INTERFACE.
                format('port-channel', '470', 267) +
            '[\x00-\x7f]+' +
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('port-channel', '470', '267')),
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('port-channel', '470', '265,267'),
            test_cisco_nexus_base.RESULT_ADD_VLAN.format('265,267'),
            (test_cisco_nexus_base.RESULT_ADD_NATIVE_INTERFACE.
                format('port-channel', '470', 267) +
            '[\x00-\x7f]+' +
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('port-channel', '470', '267')),
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('port-channel', '470', '265,267'),
            test_cisco_nexus_base.RESULT_ADD_VLAN.format('265,267')]),

    }


class TestCiscoNexusBaremetalReplay(
    test_cisco_nexus_base.TestCiscoNexusReplayBase):

    """Unit tests for Replay of Cisco ML2 Nexus data."""

    baremetal_profile = {
        "local_link_information": [
            {
                "port_id": test_cisco_nexus_base.NEXUS_BAREMETAL_PORT_1,
                "switch_info": {
                    "switch_ip": test_cisco_nexus_base.NEXUS_IP_ADDRESS_1,
                },
            },
        ]
    }

    baremetal_profile2 = {
        "local_link_information": [
            {
                "port_id": test_cisco_nexus_base.NEXUS_BAREMETAL_PORT_2,
                "switch_info": {
                    "switch_ip": test_cisco_nexus_base.NEXUS_IP_ADDRESS_1,
                },
            },
        ]
    }

    baremetal_profile_vPC = {
        "local_link_information": [
            {
                "port_id": test_cisco_nexus_base.NEXUS_BAREMETAL_PORT_1,
                "switch_info": {
                    "switch_ip": test_cisco_nexus_base.NEXUS_IP_ADDRESS_1,
                },
            },
            {
                "port_id": test_cisco_nexus_base.NEXUS_BAREMETAL_PORT_2,
                "switch_info": {
                    "switch_ip": test_cisco_nexus_base.NEXUS_IP_ADDRESS_2,
                },
            },
        ]
    }

    test_configs = {
        'test_replay_unique1':
            test_cisco_nexus_base.TestCiscoNexusBase.TestConfigObj(
                None,
                test_cisco_nexus_base.HOST_NAME_UNUSED,
                test_cisco_nexus_base.NEXUS_BAREMETAL_PORT_1,
                test_cisco_nexus_base.INSTANCE_1,
                test_cisco_nexus_base.VLAN_ID_1,
                test_cisco_nexus_base.NO_VXLAN_ID,
                None,
                test_cisco_nexus_base.DEVICE_OWNER_BAREMETAL,
                baremetal_profile,
                test_cisco_nexus_base.HOST_NAME_Baremetal + '1',
                test_cisco_nexus_base.BAREMETAL_VNIC),
        'test_replay_unique2':
            test_cisco_nexus_base.TestCiscoNexusBase.TestConfigObj(
                None,
                test_cisco_nexus_base.HOST_NAME_UNUSED,
                test_cisco_nexus_base.NEXUS_BAREMETAL_PORT_2,
                test_cisco_nexus_base.INSTANCE_2,
                test_cisco_nexus_base.VLAN_ID_2,
                test_cisco_nexus_base.NO_VXLAN_ID,
                None,
                test_cisco_nexus_base.DEVICE_OWNER_BAREMETAL,
                baremetal_profile2,
                test_cisco_nexus_base.HOST_NAME_Baremetal + '1',
                test_cisco_nexus_base.BAREMETAL_VNIC),
        'test_replay_unique_vPC':
            test_cisco_nexus_base.TestCiscoNexusBase.TestConfigObj(
                None,
                test_cisco_nexus_base.HOST_NAME_UNUSED,
                test_cisco_nexus_base.NEXUS_BAREMETAL_PORT_1,
                test_cisco_nexus_base.INSTANCE_1,
                test_cisco_nexus_base.VLAN_ID_1,
                test_cisco_nexus_base.NO_VXLAN_ID,
                None,
                test_cisco_nexus_base.DEVICE_OWNER_BAREMETAL,
                baremetal_profile_vPC,
                test_cisco_nexus_base.HOST_NAME_Baremetal + '1',
                test_cisco_nexus_base.BAREMETAL_VNIC),
        'test_config_vm':
            test_cisco_nexus_base.TestCiscoNexusBase.TestConfigObj(
                test_cisco_nexus_base.NEXUS_IP_ADDRESS_1,
                test_cisco_nexus_base.HOST_NAME_Baremetal + '1',
                test_cisco_nexus_base.NEXUS_BAREMETAL_PORT_1,
                test_cisco_nexus_base.INSTANCE_2,
                test_cisco_nexus_base.VLAN_ID_2,
                test_cisco_nexus_base.NO_VXLAN_ID,
                None,
                test_cisco_nexus_base.DEVICE_OWNER_COMPUTE,
                {},
                None,
                test_cisco_nexus_base.NORMAL_VNIC),
    }

    test_configs = collections.OrderedDict(sorted(test_configs.items()))

    def setUp(self):
        """Sets up mock ncclient, and switch and credentials dictionaries."""

        cfg.CONF.set_override('nexus_driver', 'ncclient', 'ml2_cisco')
        cfg.CONF.set_override('never_cache_ssh_connection', False, 'ml2_cisco')
        super(TestCiscoNexusBaremetalReplay, self).setUp()
        self.results = TestCiscoNexusBaremetalReplayResults()

    def _init_port_channel(self, ch_grp, which=1):

        # with Baremetal config when enet interface associated to port-channel,
        # the port-channel interface is configured instead.  This config
        # causes this to happen.
        data_xml = {'connect.return_value.get.return_value.data_xml':
                    'switchport trunk allowed vlan none\n'
                    'channel-group ' + str(ch_grp) + ' mode active'}
        self.mock_ncclient.configure_mock(**data_xml)

    def test_replay_unique_ethernet_ports(self):
        """Provides replay data and result data for unique ports. """

        first_add = {
            'driver_results': self.results.get_test_results(
                'driver_result_unique_eth_add1'),
            'nbr_db_entries': 1}
        second_add = {
            'driver_results': self.results.get_test_results(
                'driver_result_unique_eth_add2'),
            'nbr_db_entries': 2}
        first_del = {
            'driver_results': self.results.get_test_results(
                'driver_result_unique_eth_del1'),
            'nbr_db_entries': 1,
            'nbr_db_mappings': 0}
        second_del = {
            'driver_results': self.results.get_test_results(
                'driver_result_unique_eth_del2'),
            'nbr_db_entries': 0}

        self._process_replay(
            'test_replay_unique1',
            'test_replay_unique2',
            self.results.get_test_results(
                'driver_result_unique_eth_init'),
            first_add,
            second_add,
            self.results.get_test_results(
                'driver_result_unique_2if_replay'),
            first_del,
            second_del)

    def test_replay_unique_ethernet_port_and_vm(self):
        """Provides replay data and result data for unique ports. """

        first_add = {
            'driver_results': self.results.get_test_results(
                'driver_result_unique_eth_add1'),
            'nbr_db_entries': 1}
        second_add = {
            'driver_results': self.results.get_test_results(
                'driver_result_unique_eth_add_vm'),
            'nbr_db_entries': 2}
        first_del = {
            'driver_results': self.results.get_test_results(
                'driver_result_unique_eth_del_vm'),
            'nbr_db_entries': 1}
        second_del = {
            'driver_results': self.results.get_test_results(
                'driver_result_unique_eth_del2'),
            'nbr_db_entries': 0}

        self._process_replay(
            'test_replay_unique1',
            'test_config_vm',
            self.results.get_test_results(
                'driver_result_unique_eth_init'),
            first_add,
            second_add,
            self.results.get_test_results(
                'driver_result_unique_2vlan_replay'),
            first_del,
            second_del)

    def test_replay_unique_vPC_ports(self):
        """Provides replay data and result data for unique ports. """

        first_add = {
            'driver_results': self.results.get_test_results(
                'driver_result_unique_vPC_2switch_add1'),
            'nbr_db_entries': 2}
        second_del = {
            'driver_results': self.results.get_test_results(
                'driver_result_unique_vPC_2switch_del1'),
            'nbr_db_entries': 0}

        self._init_port_channel(469)

        self._process_replay(
            'test_replay_unique_vPC',
            None,
            [],
            first_add,
            None,
            self.results.get_test_results(
                'driver_result_unique_vPC_2if_replay'),
            None,
            second_del)

    def test_replay_unique_vPC_ports_and_vm(self):
        """Provides replay data and result data for unique ports. """

        first_add = {
            'driver_results': self.results.get_test_results(
                'driver_result_unique_vPC470_add1'),
            'nbr_db_entries': 2}
        second_add = {
            'driver_results': self.results.get_test_results(
                'driver_result_unique_vPC470_add2'),
            'nbr_db_entries': 4}
        first_del = {
            'driver_results': self.results.get_test_results(
                'driver_result_unique_vPC470_del1'),
            'nbr_db_entries': 2}
        second_del = {
            'driver_results': self.results.get_test_results(
                'driver_result_unique_vPC470_del2'),
            'nbr_db_entries': 0}

        self._init_port_channel(470)

        self._process_replay(
            'test_replay_unique_vPC',
            'test_config_vm',
            [],
            first_add,
            second_add,
            self.results.get_test_results(
                'driver_result_unique_vPC470_2vlan_replay'),
            first_del,
            second_del)

    def test_replay_unique_vPC_ports_chg_vPC_nbr(self):
        """Persist with learned channel group even if it changed."""

        def replay_init():
            # This causes port-channel 469 will persist instead.
            # We will not relearn
            self._init_port_channel(470)

        first_add = {
            'driver_results': self.results.get_test_results(
                'driver_result_unique_vPC_2switch_add1'),
            'nbr_db_entries': 2}
        second_del = {
            'driver_results': self.results.get_test_results(
                'driver_result_unique_vPC_2switch_del1'),
            'nbr_db_entries': 0}

        self._init_port_channel(469)

        self._process_replay(
            'test_replay_unique_vPC',
            None,
            [],
            first_add,
            None,
            self.results.get_test_results(
                'driver_result_unique_vPC_2if_replay'),
            None,
            second_del,
            replay_init)

    def test_replay_unique_vPC_ports_chg_to_enet(self):
        """Persist with learned channel group even if it was removed."""

        def replay_init():
            # This causes port-channel to get replaced with enet
            # by eliminating channel-group config from enet config.
            if cfg.CONF.ml2_cisco.nexus_driver == 'restapi':
                self.restapi_mock_init()
            else:
                self.mock_init()

        first_add = {
            'driver_results': self.results.get_test_results(
                'driver_result_unique_vPC_2switch_add1'),
            'nbr_db_entries': 2}
        second_del = {
            'driver_results': self.results.get_test_results(
                'driver_result_unique_vPC_2switch_del1'),
            'nbr_db_entries': 0}

        self._init_port_channel(469)

        self._process_replay(
            'test_replay_unique_vPC',
            None,
            [],
            first_add,
            None,
            self.results.get_test_results(
                'driver_result_unique_vPC_2if_replay'),
            None,
            second_del,
            replay_init)


class TestCiscoNexusNonCachedSshReplay(
    test_cisco_nexus_base.TestCiscoNexusReplayBase):
    """Unit tests for Replay of Cisco ML2 Nexus data."""

    # Testing new default of True for config var 'never_cache_ssh_connection'

    test_configs = {
        'test_replay_unique1':
            test_cisco_nexus_base.TestCiscoNexusBase.TestConfigObj(
                test_cisco_nexus_base.NEXUS_IP_ADDRESS_1,
                RP_HOST_NAME_1,
                test_cisco_nexus_base.NEXUS_PORT_1,
                test_cisco_nexus_base.INSTANCE_1,
                test_cisco_nexus_base.VLAN_ID_1,
                test_cisco_nexus_base.NO_VXLAN_ID,
                None,
                test_cisco_nexus_base.DEVICE_OWNER_COMPUTE,
                {},
                None,
                test_cisco_nexus_base.NORMAL_VNIC),
    }

    def setUp(self):
        cfg.CONF.set_override('nexus_driver', 'ncclient', 'ml2_cisco')
        super(TestCiscoNexusNonCachedSshReplay, self).setUp()
        self.mock_ncclient.reset_mock()

    def test_basic_replay_NonCacheSsh(self):
        """Basic none cached ssh replay test."""

        tmp_cfg = self.TestConfigObj(
            test_cisco_nexus_base.NEXUS_IP_ADDRESS_1,
            RP_HOST_NAME_1,
            test_cisco_nexus_base.NEXUS_PORT_1,
            test_cisco_nexus_base.INSTANCE_1,
            test_cisco_nexus_base.VLAN_ID_1,
            test_cisco_nexus_base.NO_VXLAN_ID,
            None,
            test_cisco_nexus_base.DEVICE_OWNER_COMPUTE,
            {},
            None,
            test_cisco_nexus_base.NORMAL_VNIC)

        self._cisco_mech_driver.set_switch_ip_and_active_state(
            tmp_cfg.nexus_ip_addr, const.SWITCH_ACTIVE)

        self.mock_ncclient.reset_mock()

        # Create a batch of port entries with unique vlans
        num_vlans = 10
        for x in range(num_vlans):
            instance_id = test_cisco_nexus_base.INSTANCE_1 + '-' + str(x)
            new_cfg = tmp_cfg._replace(
                          vlan_id=test_cisco_nexus_base.VLAN_ID_1 + x,
                          instance_id=instance_id)
            self._create_port(new_cfg)

        self.assertEqual(20, self.mock_ncclient.connect.call_count)
        self.assertEqual(20,
            self.mock_ncclient.connect.return_value.
            close_session.call_count)

        self.mock_ncclient.reset_mock()

        # Put it back to inactive state
        self._cisco_mech_driver.set_switch_ip_and_active_state(
            tmp_cfg.nexus_ip_addr, const.SWITCH_INACTIVE)

        self._cfg_monitor.check_connections()

        self.assertEqual(2, self.mock_ncclient.connect.call_count)
        self.assertEqual(2,
            self.mock_ncclient.connect.return_value.
            close_session.call_count)



# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
from datetime import datetime
import mock
import time

from neutron.api.v2 import attributes
from neutron.common import test_lib
from neutron.db import agents_db
from neutron.extensions import agent
from neutron.tests.unit.extensions import test_l3
from novaclient import exceptions as nova_exc
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils
from oslo_utils import importutils
from oslo_utils import timeutils
from oslo_utils import uuidutils

import networking_cisco
from networking_cisco._i18n import _LE
from networking_cisco import backwards_compatibility as bc
from networking_cisco import plugins
from networking_cisco.plugins.cisco.common import cisco_constants
from networking_cisco.plugins.cisco.db.device_manager import (
    hosting_device_manager_db)
from networking_cisco.plugins.cisco.db.scheduler import cfg_agentschedulers_db
from networking_cisco.plugins.cisco.device_manager.rpc import (
    devices_cfgagent_rpc_cb)
from networking_cisco.plugins.cisco.device_manager.rpc import (
    devmgr_rpc_cfgagent_api)
from networking_cisco.plugins.cisco.extensions import (
    ciscocfgagentscheduler as cfgagtscheduler)
from networking_cisco.plugins.cisco.extensions import (
    ciscohostingdevicemanager as ciscodevmgr)


LOG = logging.getLogger(__name__)


_uuid = uuidutils.generate_uuid
ISO8601_TIME_FORMAT = bc.constants.ISO8601_TIME_FORMAT

CORE_PLUGIN_KLASS = (
    'networking_cisco.tests.unit.cisco.device_manager'
    '.device_manager_test_support.TestCorePlugin')
extensions_path = (':' + plugins.__path__[0] + '/cisco/extensions')

L3_CFG_HOST_A = 'host_a'
L3_CFG_HOST_B = 'host_b'
L3_CFG_HOST_C = 'host_c'


class DeviceManagerTestSupportMixin(object):

    @property
    def _core_plugin(self):
        return bc.get_plugin()

    def _mock_l3_admin_tenant(self):
        # Mock l3 admin tenant
        self.tenant_id_fcn_p = mock.patch(
            'networking_cisco.plugins.cisco.db.device_manager.'
            'hosting_device_manager_db.HostingDeviceManagerMixin.l3_tenant_id')
        self.tenant_id_fcn = self.tenant_id_fcn_p.start()
        self.tenant_id_fcn.return_value = "L3AdminTenantId"

    def _create_mgmt_nw_for_tests(self, fmt):
        self._mgmt_nw = self._make_network(fmt,
                                           cfg.CONF.general.management_network,
                                           True, tenant_id="L3AdminTenantId",
                                           shared=False)
        self._mgmt_subnet = self._make_subnet(fmt, self._mgmt_nw,
                                              "10.0.100.1", "10.0.100.0/24",
                                              ip_version=4)

    def _remove_mgmt_nw_for_tests(self):
        q_p = "network_id=%s" % self._mgmt_nw['network']['id']
        subnets = self._list('subnets', query_params=q_p)
        if subnets:
            for p in self._list('ports', query_params=q_p).get('ports'):
                self._delete('ports', p['id'])
            self._delete('subnets', self._mgmt_subnet['subnet']['id'])
            self._delete('networks', self._mgmt_nw['network']['id'])
        hosting_device_manager_db.HostingDeviceManagerMixin._mgmt_nw_uuid = (
            None)
        hosting_device_manager_db.HostingDeviceManagerMixin._mgmt_subnet_uuid\
            = None

    # Function used to mock novaclient services list
    def _novaclient_services_list(self, all=True):
        services = set(['nova-conductor', 'nova-cert', 'nova-scheduler',
                        'nova-compute'])
        full_list = [FakeResource(binary=res) for res in services]
        _all = all

        def response():
            if _all:
                return full_list
            else:
                return full_list[2:]
        return response

    # Function used to mock novaclient servers create
    def _novaclient_servers_create(self, instance_name, image_id, flavor_id,
                                   nics, files, config_drive):
        fake_vm = FakeResource()
        for nic in nics:
            p_dict = {'port': {'device_id': fake_vm.id,
                               'device_owner': 'nova'}}
            self._core_plugin.update_port(bc.context.get_admin_context(),
                                          nic['port-id'], p_dict)
        return fake_vm

    # Function used to mock novaclient servers delete
    def _novaclient_servers_delete(self, vm_id):
        q_p = "device_id=%s" % vm_id
        ports = self._list('ports', query_params=q_p)
        for port in ports.get('ports', []):
            try:
                self._delete('ports', port['id'])
            except Exception as e:
                with excutils.save_and_reraise_exception(reraise=False):
                    LOG.error(_LE('Failed to delete port %(p_id)s for vm '
                                  'instance %(v_id)s due to %(err)s'),
                              {'p_id': port['id'], 'v_id': vm_id, 'err': e})
                    raise nova_exc.InternalServerError()

    def _mock_svc_vm_create_delete(self, plugin):
        # Mock novaclient methods for creation/deletion of service VMs
        mock.patch(
            'networking_cisco.plugins.cisco.device_manager.service_vm_lib.'
            'n_utils.find_resource',
            lambda *args, **kw: FakeResource()).start()
        self._nclient_services_mock = mock.MagicMock()
        self._nclient_services_mock.list = self._novaclient_services_list()
        mock.patch.object(plugin._svc_vm_mgr_obj._nclient, 'services',
                          self._nclient_services_mock).start()
        nclient_servers_mock = mock.MagicMock()
        nclient_servers_mock.create = self._novaclient_servers_create
        nclient_servers_mock.delete = self._novaclient_servers_delete
        mock.patch.object(plugin._svc_vm_mgr_obj._nclient, 'servers',
                          nclient_servers_mock).start()

    def _mock_dispatch_pool_maintenance(self):
        # Mock creation/deletion of service VMs
        dispatch_pool_maintenance_job_fcn_p = mock.patch(
            'networking_cisco.plugins.cisco.db.device_manager.'
            'hosting_device_manager_db.HostingDeviceManagerMixin.'
            '_dispatch_pool_maintenance_job')
        dispatch_pool_maintenance_job_fcn_p .start()

    def _mock_eventlet_greenpool_spawn_n(self):
        # Mock GreenPool's spawn_n to execute the specified function directly
        self._greenpool_mock = mock.MagicMock()
        self._greenpool_mock.return_value.spawn_n = (
            lambda f, *args, **kwargs: f(*args, **kwargs))
        _eventlet_greenpool_fcn_p = mock.patch(
            'networking_cisco.plugins.cisco.db.device_manager.'
            'hosting_device_manager_db.eventlet.GreenPool',
            self._greenpool_mock)
        _eventlet_greenpool_fcn_p.start()

    def _test_remove_all_hosting_devices(self):
        """Removes all hosting devices created during a test."""
        devmgr = bc.get_plugin(cisco_constants.DEVICE_MANAGER)
        context = bc.context.get_admin_context()
        devmgr.delete_all_hosting_devices(context, True)

    def _get_fake_resource(self, tenant_id=None, id=None):
        return {'id': id or _uuid(),
                'tenant_id': tenant_id or _uuid()}

    def _get_test_context(self, user_id=None, tenant_id=None, is_admin=False):
        return bc.context.Context(user_id, tenant_id, is_admin)

    def _mock_cfg_agent_notifier(self, plugin):
        # Mock notifications to l3 agent and Cisco config agent
        self._l3_agent_mock = mock.MagicMock()
        self._cfg_agent_mock = mock.MagicMock()
        self._l3_cfg_agent_mock = mock.MagicMock()
        plugin.agent_notifiers = {
            bc.constants.AGENT_TYPE_L3: self._l3_agent_mock,
            cisco_constants.AGENT_TYPE_CFG: self._cfg_agent_mock,
            cisco_constants.AGENT_TYPE_L3_CFG: self._l3_cfg_agent_mock}

    def _define_keystone_authtoken(self):
        test_opts = [
            cfg.StrOpt('auth_url', default='http://localhost:35357/v2.0/'),
            cfg.StrOpt('identity_uri', default='http://localhost:5000'),
            #cfg.StrOpt('admin_user', default='neutron'),
            cfg.StrOpt('username', default='neutron'),
            #cfg.StrOpt('admin_password', default='secrete'),
            cfg.StrOpt('password', default='secrete'),
            cfg.StrOpt('project_name', default='service'),
            cfg.StrOpt('project_domain_id', default='default'),
            cfg.StrOpt('user_domain_id', default='default')]
        cfg.CONF.register_opts(test_opts, 'keystone_authtoken')

    def _add_device_manager_plugin_ini_file(self):
        # includes config files for device manager service plugin
        cfg_file = (
            networking_cisco.__path__[0] +
            '/tests/unit/cisco/etc/cisco_device_manager_plugin.ini')
        if 'config_files' in test_lib.test_config:
            test_lib.test_config['config_files'].append(cfg_file)
        else:
            test_lib.test_config['config_files'] = [cfg_file]

    def _register_cfg_agent_states(self, host_a_active=True,
                                   host_b_active=False,
                                   host_c_active=False):
        """Register zero, one, two, or three L3 config agents."""
        l3_cfg_host_a = {
            'binary': 'neutron-cisco-cfg-agent',
            'host': L3_CFG_HOST_A,
            'topic': cisco_constants.CFG_AGENT,
            'configurations': {
                'service_agents': [cisco_constants.AGENT_TYPE_L3_CFG],
                'total routers': 0,
                'total ex_gw_ports': 0,
                'total interfaces': 0,
                'total floating_ips': 0,
                'hosting_devices': 0,
                'non_responding_hosting_devices': {}},
            'local_time': str(timeutils.utcnow()),
            'agent_type': cisco_constants.AGENT_TYPE_CFG}
        agent_callback = agents_db.AgentExtRpcCallback()
        dev_mgr_callback = devices_cfgagent_rpc_cb.DeviceMgrCfgRpcCallback(
            bc.get_plugin(cisco_constants.DEVICE_MANAGER))
        if host_a_active is True:
            agent_callback.report_state(
                self.adminContext,
                agent_state={'agent_state': l3_cfg_host_a},
                time=datetime.utcnow().strftime(ISO8601_TIME_FORMAT))
            dev_mgr_callback.register_for_duty(self.adminContext,
                                               L3_CFG_HOST_A)
        if host_b_active is True:
            l3_cfg_host_b = copy.deepcopy(l3_cfg_host_a)
            l3_cfg_host_b['host'] = L3_CFG_HOST_B
            l3_cfg_host_b['local_time'] = str(timeutils.utcnow()),
            agent_callback.report_state(
                self.adminContext, agent_state={'agent_state': l3_cfg_host_b},
                time=datetime.utcnow().strftime(ISO8601_TIME_FORMAT))
            dev_mgr_callback.register_for_duty(self.adminContext,
                                               L3_CFG_HOST_B)
        if host_c_active is True:
            l3_cfg_host_c = copy.deepcopy(l3_cfg_host_a)
            l3_cfg_host_c['host'] = L3_CFG_HOST_C
            l3_cfg_host_c['local_time'] = str(timeutils.utcnow()),
            agent_callback.report_state(
                self.adminContext, agent_state={'agent_state': l3_cfg_host_c},
                time=datetime.utcnow().strftime(ISO8601_TIME_FORMAT))
            dev_mgr_callback.register_for_duty(self.adminContext,
                                               L3_CFG_HOST_B)


class TestDeviceManagerExtensionManager(object):

    def get_resources(self):
        # NOTE(bobmel): In rare cases, the device manager service plugin has
        # not been registered completely when the resources for it are
        # processed here. That will result in an intermittent KeyError. We
        # handle this problem by a number of retries.
        max_attempts = 5
        for i in range(max_attempts):
            try:
                res = ciscodevmgr.Ciscohostingdevicemanager.get_resources()
                break
            except KeyError as e:
                if i < max_attempts - 1:
                    LOG.debug('Retrying get_resources due to: %s', e)
                    time.sleep(2)
                else:
                    raise
        # add agent resource
        for item in agent.Agent.get_resources():
            res.append(item)
        # add hosting device to cfg agent scheduler resources
        for item in cfgagtscheduler.Ciscocfgagentscheduler.get_resources():
            res.append(item)
        # Add the resources to the global attribute map
        # This is done here as the setup process won't
        # initialize the main API router which extends
        # the global attribute map
        attributes.RESOURCE_ATTRIBUTE_MAP.update(
            ciscodevmgr.RESOURCE_ATTRIBUTE_MAP)
        return res

    def get_actions(self):
        return []

    def get_request_extensions(self):
        return []


# A core plugin supporting Cisco device manager and hosting device to cfg
# agent scheduling functionality
class TestCorePlugin(test_l3.TestNoL3NatPlugin,
                     cfg_agentschedulers_db.CfgAgentSchedulerDbMixin,
                     hosting_device_manager_db.HostingDeviceManagerMixin):

    supported_extension_aliases = [
        "agent", "external-net",
        cfgagtscheduler.CFG_AGENT_SCHEDULER_ALIAS,
        ciscodevmgr.HOSTING_DEVICE_MANAGER_ALIAS]

    def __init__(self):
        super(TestCorePlugin, self).__init__()
        self.cfg_agent_scheduler = importutils.import_object(
            cfg.CONF.general.configuration_agent_scheduler_driver)
        self.agent_notifiers[cisco_constants.AGENT_TYPE_CFG] = (
            devmgr_rpc_cfgagent_api.DeviceMgrCfgAgentNotifyAPI(self))

    def cleanup_after_test(self):
        """This function should be called in the TearDown() function of
        test classes that use the plugin.

        Reset all class variables to their default values.
        This is needed to avoid tests to pollute subsequent tests.
        """
        TestCorePlugin._l3_tenant_uuid = None
        TestCorePlugin._mgmt_nw_uuid = None
        TestCorePlugin._mgmt_subnet_uuid = None
        TestCorePlugin._mgmt_sec_grp_id = None
        TestCorePlugin._credentials = {}
        TestCorePlugin._plugging_drivers = {}
        TestCorePlugin._hosting_device_drivers = {}
        TestCorePlugin._hosting_device_locks = {}
        TestCorePlugin._cfgagent_scheduler = None
        TestCorePlugin._cfg_agent_statuses = {}
        TestCorePlugin._svc_vm_mgr_obj = None
        TestCorePlugin._nova_running = False


# Used to fake Glance images, Nova VMs and Nova services
class FakeResource(object):
    def __init__(self, id=None, enabled='enabled', state='up', binary=None):
        self.id = id or _uuid()
        self.status = enabled
        self.state = state
        self.binary = binary



# Copyright 2018 Cisco Systems, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Remove the N1kv mechanism driver

Revision ID: c52284b920fd
Revises: 53f08de0523f
Create Date: 2018-01-16 16:48:18.526764

"""

# revision identifiers, used by Alembic.
revision = 'c52284b920fd'
down_revision = '53f08de0523f'

from alembic import op


def upgrade():
    op.drop_table('cisco_ml2_n1kv_policy_profiles')
    op.drop_table('cisco_ml2_n1kv_port_bindings')
    op.drop_table('cisco_ml2_n1kv_network_bindings')
    op.drop_table('cisco_ml2_n1kv_vlan_allocations')
    op.drop_table('cisco_ml2_n1kv_vxlan_allocations')
    op.drop_table('cisco_ml2_n1kv_profile_bindings')
    op.drop_table('cisco_ml2_n1kv_network_profiles')
